/***************************************************************
 * LangstonScreen.cpp
 * written by Steve Sprang
 * Copyright  1996
 *
 * Portions of this code were taken from Be's BWindowScreen
 * demo program: FullKaleid. Specifically, from KalWindow.cpp
 **************************************************************/
 
#include "LangstonScreen.h"

// in main.cpp
extern sem_id screen;

/***************************************************************
 * distance() : returns the distance between two points
 **************************************************************/
double distance(long x1, long y1, long x2, long y2)
{
	return(sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)));
}

/***************************************************************
 * LangstonEntry()
 **************************************************************/
static long LangstonEntry(void *param)
{
	LangstonScreen 		*window;

	// Get a pointer to the WindowScreen object.
	window = (LangstonScreen*)param;

	// init stuff
	window->Reset();
	// start the ant
	window->Langston();
	
	return 1;
}

/***************************************************************
 * LangstonScreen::LangstonScreen()
 **************************************************************/
LangstonScreen::LangstonScreen(const char *title, ulong space)
	: BWindowScreen(title, space)
{
	rgb_color    palette[256];

	// preset the drawing thread pointer to 0.
	draw = NULL;
	
	// Abort everything else if the constructor failed.	
	if (Error() == B_ERROR) return;

	// Change the color_map. 
	// black at index 0, white at index 1
	palette[0].red = palette[0].green = palette[0].blue = 0;
	palette[1].red = palette[1].green = palette[1].blue = 255;
	
	SetColorList(palette);
	
	// get gfx card hooks
	GetHooks();
	
	// allocate huge 2D array of booleans
	arena = new bool*[width+1];
	for(int i = 0; i < width+1; i++) {
		arena[i] = new bool[height+1];
	}
	
	// Spawn and resume the drawing thread
	draw = spawn_thread(LangstonEntry,"langston draw",B_DISPLAY_PRIORITY,(void*)this);
	resume_thread(draw);
}

/***************************************************************
 * LangstonScreen::Langston()
 **************************************************************/
void LangstonScreen::Langston()
{
	int col, count = 0;
	int pos = 0;
	
	// Because "right" and "left" are relative to current
	// direction, this array of turning values is necessary.
	// Stepping through by two tells you how to turn right;
	// stepping through by negative two tells you how to turn
	// left (based on your current pos in the list).
	int dir[8] = {1,0,0,-1,-1,0,0,1};
	
	// get gfx card hooks
	GetHooks();
	
	// start in the center
	x = width / 2;
	y = height / 2;
	
	// take control
	acquire_sem(screen);
	
	while(true) {
		
		if(arena[x][y]) {
			// the pixel is on, so turn right
			pos += 2;
			col = 1;
		}
		else {
			// the pixel is off, so turn left
			pos -= 2;
			col = 0;
		}
		
		// make sure we wrap correctly within the dir array
		if(pos == 8) {
			pos = 0;
		}
		if(pos == -2) {
			pos = 6;
		}
		
		// update x and y based on the direction we want to turn
		x += dir[pos];
		y += dir[pos+1];
		
		// make sure we wrap correctly around the screen
		if(x < 0) {
			x = width;
		}
		else if(x > width) {
			x = 0;
		}

		if(y < 0) {
			y = height;
		}
		else if(y > height) {
			y = 0;
		}
		
		// invert the value of this pixel
		arena[x][y] = !arena[x][y];
		
		// color the pixel appropriately
		(draw_line)(x, x, y, y, col, 0, 0, 0, 0, 0);
		
		// release control every so often
		if(count == 1024) {
			release_sem(screen);
			count = 0;
			acquire_sem(screen);
			GetHooks();
		}
		count++;
	}
}

/***************************************************************
 * LangstonScreen::GetHooks()
 **************************************************************/
void LangstonScreen::GetHooks()
{
	// According to Be:
	// Each time you release and take again the control of the screen,
	// it's better to reload all the graphics add-ons dependent hooks and
	// values (see documentation for more informations).
	draw_line = (LINE)CardHookAt(INDEX_LINE);
	fill_rect = (RECT)CardHookAt(INDEX_RECT);

	// width and height are defined last value included
	width = CardInfo()->width-1;
	height = CardInfo()->height-1;
}

/***************************************************************
 * LangstonScreen::RandomDots()
 **************************************************************/
void LangstonScreen::RandomDots()
{
	long x1, y1, cx, cy;
	
	// make the ant wait
	acquire_sem(screen);
	
	// get gfx card hooks
	GetHooks();

	// Fill everything in black.
	if (fill_rect != 0L)
		(fill_rect)(0, 0, width, height, 0);
	
	// find the center	
	cx = width / 2;
	cy = height / 2;
	
	// generate a circle of random dots for the ant to play with
	for(int i = 0; i < 2000; i++) {
		// generate a point inside a (2*kRADIUS) x (2*kRADIUS) square
		x1 = rand() % (2*kRADIUS);
		y1 = rand() % (2*kRADIUS);
		
		// translate to the center
		x1 += (cx - kRADIUS);
		y1 += (cy - kRADIUS);
			
		// draw only if it's inside the circle of radius kRADIUS
		if(distance(x1, y1, cx, cy) < kRADIUS) {
			(draw_line)(x1, x1, y1, y1, 1, 0, 0, 0, 0, 0);
			arena[x1][y1] = TRUE;
		}
	}
	
	// let the ant go
	release_sem(screen);
}	


/***************************************************************
 * LangstonScreen::Reset()
 **************************************************************/
void LangstonScreen::Reset()
{
	// init huge 2D array of booleans
	for(int i = 0; i < width+1; i++) {
		for(int n = 0; n < height+1; n++) {
			arena[i][n] = FALSE;
		}
	}
	
	// reset ant position
	x = width / 2;
	y = height / 2;
	
	// create circle of random dots
	RandomDots();
}


/***************************************************************
 * LangstonScreen::~LangstonScreen()
 **************************************************************/
LangstonScreen::~LangstonScreen()
{
	// clean up huge 2D array of booleans
	for(int i = 0; i < width; i++) {
		delete arena[i];
	}
	delete arena;
}

/***************************************************************
 * LangstonScreen::ScreenConnected()
 **************************************************************/
void LangstonScreen::ScreenConnected(bool active)
{
	if (active == TRUE)
		// Release the semaphore to allow the drawing thread to 
		// access the screen.
		release_sem(screen);
	else
		// Acquire the semaphore to block the drawing thread.
		acquire_sem(screen);
}

/***************************************************************
 * LangstonScreen::Quit()
 **************************************************************/
void LangstonScreen::Quit()
{	
	// if the drawing thread had been created...
	if (draw != 0L) {
		// ...disconnect the screen control to block that thread 
		//    in a clean state...
		Disconnect();
		// ...and then kill it !
		kill_thread(draw);
	}
	// Quit the WindowScreen.
	BWindowScreen::Quit();
}

/***************************************************************
 * LangstonScreen::MessageReceived()
 **************************************************************/
void LangstonScreen::MessageReceived(BMessage *msg)
{
	switch(msg->what) {
		
		case B_KEY_DOWN:
		{
			long key = msg->FindLong("key");
			long aChar = msg->FindLong("char");
			
			switch (aChar) {
				case B_ESCAPE:
					be_app->PostMessage(B_QUIT_REQUESTED);
					break;
				case B_SPACE:
					// start a new ant
					Reset();
					break;
			}
		}
	}
}